"""Model252 module."""

from dataclasses import dataclass, field
from typing import ClassVar

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from ..hlm import ModelForcings, ModelParameters, ModelStates, NodeParameters
from ..hlm_bmi import BmiFields, HlmBmi

DAYS_IN_MONTH = 30
SECONDS_IN_HOUR = 3600.0
SECONDS_IN_DAY = 86400.0
MILLIMETER_IN_METER = 1000.0

S_R = 1.0  # m. Ref val


@dataclass
class Model252NodeParameters(NodeParameters):
    """Model252 node parameters."""

    a_i: float
    l_i: float
    a_h: float
    invtau: float
    k_2: float
    k_i: float


@dataclass
class Model252States(ModelStates):
    """Model252 states."""

    fields: ClassVar[tuple] = ("q", "s_p", "s_t", "s_s")

    q: float = 1.0e-6  # Channel discharge [m**3 s**-1]
    s_p: float = 0.0  # Water ponded on hillslope surface [m]
    s_t: float = 0.0  # Effective water depth in the top soil layer [m]
    s_s: float = 0.0  # Effective water depth in hillslope subsurface [m]


@dataclass
class Model252Parameters(
    ModelParameters
):  # pylint:disable=too-many-instance-attributes
    """Model252 global parameters."""

    v_0: float = 0.33  # m s-1
    lambda_1: float = 0.2
    lambda_2: float = -0.1
    v_h: float = 0.02  # m s-1
    k_3: float = 2.0425e-6 / 60.0  # s-1
    k_i_factor: float = 0.02
    h_b: float = 0.5  # m
    s_l: float = 0.1  # m
    a: float = 0.0
    b: float = 99.0
    exponent: float = 3.0
    v_b: float = 0.75  # m s-1


@dataclass
class Model252BmiFields(BmiFields):
    component_name = "HLM#252"
    input_var_names = (
        "atmosphere_water__precipitation_leq-volume_flux",
        "land_surface_water__potential_evaporation_volume_flux",
    )
    input_var_units = (
        "m s-1",
        "m s-1",
    )
    output_var_names = (
        "channel_discharge",
        "water_ponded_on_hillslope_surface",
        "effective_water_depth_in_the_top_soil_layer",
        "effective_water_depth_in_hillslope_subsurface",
    )
    output_var_units = (
        "m3 s-1",
        "m",
        "m",
        "m",
    )


@dataclass
class Model252Forcings(ModelForcings):
    """Model252 forcings."""

    precipitation: pd.DataFrame = field(default_factory=pd.DataFrame)
    potential_evaporation: pd.DataFrame = field(default_factory=pd.DataFrame)


@dataclass
class EvaporationFluxes:
    """Evaporation fluxes dataclass."""

    e_p: float
    e_t: float
    e_s: float


@dataclass
class StorageFluxes:
    """Storage fluxes dataclass."""

    q_pl: float
    q_pt: float
    q_ts: float
    q_sl: float


@dataclass
class Model252(HlmBmi[Model252Parameters, Model252Forcings, Model252States]):
    """Model252 class."""

    parameters: Model252Parameters = field(default_factory=Model252Parameters)
    forcings_data: Model252Forcings = field(default_factory=Model252Forcings)
    initial_values: Model252States = field(default_factory=Model252States)

    bmi_fields: BmiFields = field(default_factory=Model252BmiFields)

    def equations(
        self,
        t: float,
        y: NDArray[np.float64],
        node_id: int,
        # dense_output: bool,
    ) -> NDArray[np.float64]:
        """Compute the updated derivatives using dense solutions."""

        node_param: Model252NodeParameters = self.network.nodes[node_id]["parameters"]
        states = Model252States.from_array(y)
        children_solutions = self.get_children_solutions(node_id, t)

        precipitation = self.get_input(self.forcings_data.precipitation, node_id, t)
        evaporation_fluxes = self.get_evaporation_fluxes(t, node_id, states)
        storage_fluxes = self.get_storage_fluxes(states, node_param)

        discharge = self.get_discharge(
            states,
            node_param,
            storage_fluxes,
            children_solutions,
        )

        ponded = (
            precipitation
            - storage_fluxes.q_pl
            - storage_fluxes.q_pt
            - evaporation_fluxes.e_p
        )
        topsoil = storage_fluxes.q_pt - storage_fluxes.q_ts - evaporation_fluxes.e_t
        subsurface = storage_fluxes.q_ts - storage_fluxes.q_sl - evaporation_fluxes.e_s

        return np.array(
            [
                discharge,
                ponded,
                topsoil,
                subsurface,
            ]
        )

    def get_evaporation_fluxes(
        self,
        t: float,
        node_id: int,
        states: Model252States,
    ) -> EvaporationFluxes:
        """Get evaporation fluxes."""

        model_param = self.parameters

        potential_evaporation = self.get_input(
            self.forcings_data.potential_evaporation, node_id, t
        )

        if potential_evaporation <= 0.0:
            return EvaporationFluxes(0.0, 0.0, 0.0)

        corr = (
            states.s_p
            + states.s_t / model_param.s_l
            + states.s_s / (model_param.h_b - model_param.s_l)
        )
        if corr <= 1e-12:
            return EvaporationFluxes(0.0, 0.0, 0.0)

        e_p = (states.s_p / S_R) * potential_evaporation / corr
        e_t = states.s_t / model_param.s_l * potential_evaporation / corr
        e_s = (
            states.s_s
            / (model_param.h_b - model_param.s_l)
            * potential_evaporation
            / corr
        )
        return EvaporationFluxes(e_p, e_t, e_s)

    def get_storage_fluxes(
        self,
        states: Model252States,
        node_param: Model252NodeParameters,
    ) -> StorageFluxes:
        """Get fluxes between storages."""

        model_param = self.parameters

        pow_term = (
            np.power(1.0 - states.s_t / model_param.s_l, model_param.exponent)
            if (1.0 - states.s_t / model_param.s_l > 0.0)
            else 0.0
        )
        k_t = node_param.k_2 * (model_param.a + model_param.b * pow_term)

        # Fluxes
        q_pl = node_param.k_2 * states.s_p  # q_{pc} = k_2 * s_p # [m/s]
        q_pt = k_t * states.s_p  # q_{pt} = k_t * s_p # [m/s]
        q_ts = node_param.k_i * states.s_t  # q_{ts} = k_i * s_t # [m/s]
        q_sl = model_param.k_3 * states.s_s  # q_{sc} = k_3 * s_s # [m/s]

        return StorageFluxes(q_pl, q_pt, q_ts, q_sl)

    def get_discharge(
        self,
        states: Model252States,
        node_param: Model252NodeParameters,
        storage_fluxes: StorageFluxes,
        children_solutions: list[tuple[NDArray[np.float64], float]],
    ) -> float:
        """Get discharge."""

        model_param = self.parameters
        # Discharge
        discharge = -states.q + node_param.a_h * (  # m3/s
            storage_fluxes.q_pl + storage_fluxes.q_sl
        )  # m2*(m/s) == m3/s

        # children_solutions[child_id] = (child_ode_result, child_flow_partition)

        for child_states, child_flow_partition in children_solutions:
            child_states_arr = Model252States.from_array(child_states)
            discharge_child = child_states_arr.q * child_flow_partition

            if discharge_child < 0:
                discharge_child = 1e-6
            discharge += discharge_child

        if states.q < 0 and model_param.lambda_1 < 1:
            return 0.0

        return node_param.invtau * np.power(states.q, model_param.lambda_1) * discharge

    def _calculate_additional_network_parameters(self) -> None:
        for node_id in self.network.nodes:
            model_parameters = self.parameters
            node_parameters = self.network.nodes[node_id]["parameters"]

            a_i = node_parameters[0]
            l_i = node_parameters[1] * 1e3  # km -> m
            a_h = node_parameters[2] * 1e6  # km2 -> m2
            # models/definitions/ConvertParams

            # Compute extra local parameters
            invtau = (
                model_parameters.v_0
                * pow(a_i, model_parameters.lambda_2)
                / ((1.0 - model_parameters.lambda_1) * l_i)
            )  # [1/s]

            k_2 = model_parameters.v_h * l_i / a_h  # [1/s]

            k_i = k_2 * model_parameters.k_i_factor  # [1/s]

            self.network.nodes[node_id]["parameters"] = Model252NodeParameters(
                a_i, l_i, a_h, invtau, k_2, k_i
            )

    def _load_forcing_series(self) -> None:
        super()._load_forcing_series()
        # Convert from mm/hr to m/s
        self.forcings_data.precipitation /= SECONDS_IN_HOUR * MILLIMETER_IN_METER
        # Convert from mm/month to m/s
        self.forcings_data.potential_evaporation /= (
            DAYS_IN_MONTH * SECONDS_IN_DAY * MILLIMETER_IN_METER
        )



"""Model254 module."""

from dataclasses import dataclass, field
from typing import ClassVar

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from ..hlm import ModelForcings, ModelParameters, ModelStates, NodeParameters
from ..hlm_bmi import BmiFields, HlmBmi

DAYS_IN_MONTH = 30
SECONDS_IN_HOUR = 3600.0
SECONDS_IN_DAY = 86400.0
MILLIMETER_IN_METER = 1000.0

S_R = 1.0  # m. Ref val


@dataclass
class Model254NodeParameters(NodeParameters):
    """Model254 node parameters."""

    a_i: float
    l_i: float
    a_h: float
    invtau: float
    k_2: float
    k_i: float


@dataclass
class Model254States(ModelStates):
    """Model254 states."""

    fields: ClassVar[tuple] = ("q", "s_p", "s_t", "s_s", "s_precip", "v_r", "q_b")

    q: float = 1.0e-6  # Channel discharge [m**3 s**-1]
    s_p: float = 0.0  # Water ponded on hillslope surface [m]
    s_t: float = 0.0  # Effective water depth in the top soil layer [m]
    s_s: float = 0.0  # Effective water depth in hillslope subsurface [m]
    s_precip: float = 0.0  # Total fallen precipitation from time 0 to t [m]
    v_r: float = 0.0  # Total flux of water from runoff from time 0 to t [m**3 s**-1]
    q_b: float = 1.0e-6  # Channel discharge from baseflow [m**3 s**-1]

    @classmethod
    def from_array(cls, arr: NDArray[np.float64]):
        """Init an instance from an array."""
        instance = super().from_array(arr)
        # Apply the special constraint for q_b
        instance.q_b = max(0.001, instance.q_b)
        return instance


@dataclass
class Model254Parameters(
    ModelParameters
):  # pylint:disable=too-many-instance-attributes
    """Model254 global parameters."""

    v_0: float = 0.33  # m s-1
    lambda_1: float = 0.2
    lambda_2: float = -0.1
    v_h: float = 0.02  # m s-1
    k_3: float = 2.0425e-6 / 60.0  # s-1
    k_i_factor: float = 0.02
    h_b: float = 0.5  # m
    s_l: float = 0.1  # m
    a: float = 0.0
    b: float = 99.0
    exponent: float = 3.0
    v_b: float = 0.75  # m s-1


class Model254BmiFields(BmiFields):
    component_name = "HLM#254"
    input_var_names = (
        "atmosphere_water__precipitation_leq-volume_flux",
        "land_surface_water__potential_evaporation_volume_flux",
    )
    input_var_units = (
        "m s-1",
        "m s-1",
    )
    output_var_names = (
        "channel_discharge",
        "water_ponded_on_hillslope_surface",
        "effective_water_depth_in_the_top_soil_layer",
        "effective_water_depth_in_hillslope_subsurface",
        "total_precipitation_from_time_0_to_t",
        "total_runoff_from_time_0_to_t",
        "channel_discharge_from_baseflow",
    )
    output_var_units = (
        "m3 s-1",
        "m",
        "m",
        "m",
        "m",
        "m",
        "m3 s-1",
    )


@dataclass
class Model254Forcings(ModelForcings):
    """Model254 forcings."""

    precipitation: pd.DataFrame = field(default_factory=pd.DataFrame)
    potential_evaporation: pd.DataFrame = field(default_factory=pd.DataFrame)


@dataclass
class EvaporationFluxes:
    """Evaporation fluxes dataclass."""

    e_p: float
    e_t: float
    e_s: float


@dataclass
class StorageFluxes:
    """Storage fluxes dataclass."""

    q_pl: float
    q_pt: float
    q_ts: float
    q_sl: float


@dataclass
class Model254(HlmBmi[Model254Parameters, Model254Forcings, Model254States]):
    """Model254 class."""

    parameters: Model254Parameters = field(default_factory=Model254Parameters)
    forcings_data: Model254Forcings = field(default_factory=Model254Forcings)
    initial_values: Model254States = field(default_factory=Model254States)

    bmi_fields: BmiFields = field(default_factory=Model254BmiFields)

    def equations(
        self,
        t: float,
        y: NDArray[np.float64],
        node_id: int,
        # dense_output: bool,
    ) -> NDArray[np.float64]:
        """Compute the updated derivatives using dense solutions."""

        node_param: Model254NodeParameters = self.network.nodes[node_id]["parameters"]
        states = Model254States.from_array(y)
        children_solutions = self.get_children_solutions(node_id, t)

        precipitation = self.get_input(self.forcings_data.precipitation, node_id, t)
        evaporation_fluxes = self.get_evaporation_fluxes(t, node_id, states)
        storage_fluxes = self.get_storage_fluxes(states, node_param)

        discharge = self.get_discharge(
            states,
            node_param,
            storage_fluxes,
            children_solutions,
        )

        ponded = (
            precipitation
            - storage_fluxes.q_pl
            - storage_fluxes.q_pt
            - evaporation_fluxes.e_p
        )
        topsoil = storage_fluxes.q_pt - storage_fluxes.q_ts - evaporation_fluxes.e_t
        subsurface = storage_fluxes.q_ts - storage_fluxes.q_sl - evaporation_fluxes.e_s

        baseflow = self.get_baseflow(
            node_param, states, children_solutions, storage_fluxes
        )

        return np.array(
            [
                discharge,
                ponded,
                topsoil,
                subsurface,
                precipitation,
                storage_fluxes.q_pl,
                baseflow,
            ]
        )

    def get_baseflow(
        self,
        node_param: Model254NodeParameters,
        states: Model254States,
        children_solutions: list[tuple[NDArray[np.float64], float]],
        storage_fluxes: StorageFluxes,
    ):
        """Get baseflow."""
        model_param = self.parameters

        base_flow = storage_fluxes.q_sl * node_param.a_h - states.q_b  # * 60.0

        for child_states, _ in children_solutions:
            child_states = Model254States.from_array(child_states)
            baseflow_child = child_states.q_b

            if baseflow_child < 0:
                baseflow_child = 1e-6
            base_flow += baseflow_child

        base_flow *= model_param.v_b / node_param.l_i
        return base_flow

    def get_evaporation_fluxes(
        self,
        t: float,
        node_id: int,
        states: Model254States,
    ) -> EvaporationFluxes:
        """Get evaporation fluxes."""

        model_param = self.parameters

        potential_evaporation = self.get_input(
            self.forcings_data.potential_evaporation, node_id, t
        )

        if potential_evaporation <= 0.0:
            return EvaporationFluxes(0.0, 0.0, 0.0)

        corr = (
            states.s_p
            + states.s_t / model_param.s_l
            + states.s_s / (model_param.h_b - model_param.s_l)
        )
        if corr <= 1e-12:
            return EvaporationFluxes(0.0, 0.0, 0.0)

        e_p = (states.s_p / S_R) * potential_evaporation / corr
        e_t = states.s_t / model_param.s_l * potential_evaporation / corr
        e_s = (
            states.s_s
            / (model_param.h_b - model_param.s_l)
            * potential_evaporation
            / corr
        )
        return EvaporationFluxes(e_p, e_t, e_s)

    def get_storage_fluxes(
        self,
        states: Model254States,
        node_param: Model254NodeParameters,
    ) -> StorageFluxes:
        """Get fluxes between storages."""

        model_param = self.parameters

        pow_term = (
            np.power(1.0 - states.s_t / model_param.s_l, model_param.exponent)
            if (1.0 - states.s_t / model_param.s_l > 0.0)
            else 0.0
        )
        k_t = node_param.k_2 * (model_param.a + model_param.b * pow_term)

        # Fluxes
        q_pl = node_param.k_2 * states.s_p  # q_{pc} = k_2 * s_p # [m/s]
        q_pt = k_t * states.s_p  # q_{pt} = k_t * s_p # [m/s]
        q_ts = node_param.k_i * states.s_t  # q_{ts} = k_i * s_t # [m/s]
        q_sl = model_param.k_3 * states.s_s  # q_{sc} = k_3 * s_s # [m/s]

        return StorageFluxes(q_pl, q_pt, q_ts, q_sl)

    def get_discharge(
        self,
        states: Model254States,
        node_param: Model254NodeParameters,
        storage_fluxes: StorageFluxes,
        children_solutions: list[tuple[NDArray[np.float64], float]],
    ) -> float:
        """Get discharge."""

        model_param = self.parameters
        # Discharge
        discharge = -states.q + node_param.a_h * (  # m3/s
            storage_fluxes.q_pl + storage_fluxes.q_sl
        )  # m2*(m/s) == m3/s

        # children_solutions[child_id] = (child_ode_result, child_flow_partition)

        for child_states, child_flow_partition in children_solutions:
            child_states_arr = Model254States.from_array(child_states)
            discharge_child = child_states_arr.q * child_flow_partition

            if discharge_child < 0:
                discharge_child = 1e-6
            discharge += discharge_child

        if states.q < 0 and model_param.lambda_1 < 1:
            return 0.0

        return node_param.invtau * np.power(states.q, model_param.lambda_1) * discharge

    def _calculate_additional_network_parameters(self) -> None:
        for node_id in self.network.nodes:
            model_parameters = self.parameters
            node_parameters = self.network.nodes[node_id]["parameters"]

            a_i = node_parameters[0]
            l_i = node_parameters[1] * 1e3  # km -> m
            a_h = node_parameters[2] * 1e6  # km2 -> m2
            # models/definitions/ConvertParams

            # Compute extra local parameters
            invtau = (
                model_parameters.v_0
                * pow(a_i, model_parameters.lambda_2)
                / ((1.0 - model_parameters.lambda_1) * l_i)
            )  # [1/s]

            k_2 = model_parameters.v_h * l_i / a_h  # [1/s]

            k_i = k_2 * model_parameters.k_i_factor  # [1/s]

            self.network.nodes[node_id]["parameters"] = Model254NodeParameters(
                a_i, l_i, a_h, invtau, k_2, k_i
            )

    def _load_forcing_series(self) -> None:
        super()._load_forcing_series()
        # Convert from mm/hr to m/s
        self.forcings_data.precipitation /= SECONDS_IN_HOUR * MILLIMETER_IN_METER
        # Convert from mm/month to m/s
        self.forcings_data.potential_evaporation /= (
            DAYS_IN_MONTH * SECONDS_IN_DAY * MILLIMETER_IN_METER
        )



"""Base HLM class."""

import logging as log
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from typing import ClassVar, Generic, TypeVar

import networkx as nx
import numpy as np
import pandas as pd
import scipy.integrate as spi
import yaml
from numpy.typing import NDArray
from tqdm.auto import tqdm

P = TypeVar("P", bound="ModelParameters")
F = TypeVar("F", bound="ModelForcings")
S = TypeVar("S", bound="ModelStates")


@dataclass
class ModelParameters:
    """Base class for model parameters."""


@dataclass
class NodeParameters:
    """Base class for node parameters."""


@dataclass
class ModelForcings:
    """Base class for forcings."""


@dataclass
class ModelStates:
    """Base class for model states."""

    fields: ClassVar[tuple] = ()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not cls.fields:
            raise NotImplementedError(f"{cls.__name__} must define 'fields' ClassVar")

    @classmethod
    def from_array(cls, arr: NDArray[np.float64]):
        """Create an instance from an array."""
        instance = cls()
        for i, field in enumerate(cls.fields):
            setattr(instance, field, arr[i])
        return instance

    def to_array(self) -> NDArray[np.float64]:
        """Convert the instance to an array."""
        return np.array(
            [getattr(self, field) for field in self.fields], dtype=np.float64
        )


@dataclass
class Hlm(ABC, Generic[P, F, S]):
    """BaseModel class."""

    parameters: P
    forcings_data: F  # TODO: load into each node?
    initial_values: S  # TODO: load into each node?

    config: dict = field(default_factory=dict)
    current_forcings: np.ndarray = field(default_factory=lambda: np.array([]))
    current_states: np.ndarray = field(default_factory=lambda: np.array([]))
    network: nx.DiGraph = field(default_factory=nx.DiGraph)
    current_time = 0.0
    flow_partitions: dict = field(default_factory=dict)  # TODO: load into each node?

    def _load_model_parameters(self) -> None:
        """Update model parameters from YAML file."""
        if (
            isinstance(self.config, dict)
            and "parameters" in self.config
            and isinstance(self.config["parameters"], dict)
        ):
            for key, value in self.config["parameters"].items():
                if hasattr(self.parameters, key):
                    log.info(
                        "Updating `parameters.%s` (old: %s, new: %s)",
                        key,
                        getattr(self.parameters, key),
                        value,
                    )
                    setattr(self.parameters, key, value)
                else:
                    log.warning("Unexpected `parameters.%s` in YAML file", key)

    def _load_initial_values_from_config(self) -> None:
        """Update initial_values from YAML file."""
        if (
            isinstance(self.config, dict)
            and "initial_values" in self.config
            and isinstance(self.config["initial_values"], dict)
        ):
            for key, value in self.config["initial_values"].items():
                if hasattr(self.initial_values, key):
                    log.info(
                        "Updating `initial_values.%s` (old: %s, new: %s)",
                        key,
                        getattr(self.initial_values, key),
                        value,
                    )
                    setattr(self.initial_values, key, value)
                else:
                    log.warning("Unexpected `initial_values.%s` in YAML file", key)

    def _load_config(self, config_file: str) -> None:
        """Load self.config from a YAML file."""

        log.info("Config file: %s", config_file)

        try:
            with open(config_file, "r", encoding="utf-8") as file:
                self.config = yaml.safe_load(file)
                log.debug(self.config)
        except (FileNotFoundError, PermissionError, yaml.YAMLError) as exc:
            log.critical("Error reading the config file: `%s`", exc)

    def _load_network_topology(self) -> None:
        """Create a DAG from a *.rvr file"""

        log.info("Network topology file: %s", self.config["rvr_file"])

        with open(self.config["rvr_file"], "r", encoding="utf-8") as file:
            lines = file.readlines()

        num_nodes = int(lines[0].strip())

        index = 1
        while index < len(lines):
            while index < len(lines) and lines[index].strip() == "":
                index += 1
            if index >= len(lines):
                break
            node_id = int(lines[index].strip())
            index += 1

            children = []  # Initialize children as empty list

            if index < len(lines) and lines[index].strip() != "":
                node_info = lines[index].strip().split()
                num_children = int(node_info[0])
                children = list(map(int, node_info[1 : num_children + 1]))
                assert num_children == len(children)
                index += 1

            self.network.add_node(node_id)

            for child_id in children:
                self.network.add_edge(node_id, child_id)

        if num_nodes != self.network.number_of_nodes():
            log.warning(
                "Number of nodes mismatch (expected: %s, actual: %s)",
                num_nodes,
                self.network.number_of_nodes(),
            )

        log.info("Loaded a %s", self.network)

    def _load_network_parameters(self) -> None:
        """Add node parameters to a DAG from a *.prm file"""
        log.info("Network parameters file: %s", self.config["prm_file"])

        # Initialize variables
        lines = []
        node_params = []
        num_nodes = 0

        try:
            with open(self.config["prm_file"], "r", encoding="utf-8") as file:
                lines = file.readlines()
        except (FileNotFoundError, PermissionError) as exc:
            log.critical(exc)
            return  # Exit early if file can't be read

        try:
            num_nodes = int(lines[0].strip())
            if num_nodes != self.network.number_of_nodes():
                log.critical("Number of nodes mismatch")
                return  # Exit early on mismatch

            index = 1
            while index < len(lines):
                while index < len(lines) and lines[index].strip() == "":
                    index += 1
                if index >= len(lines):
                    break
                node_id = int(lines[index].strip())
                index += 1
                if index < len(lines) and lines[index].strip() != "":
                    node_params = list(map(float, lines[index].strip().split()))
                    self.network.nodes[node_id]["parameters"] = node_params
                    index += 1
        except (IndexError, ValueError) as exc:
            log.critical("Invalid file: %s", exc)
            return  # Exit early on parsing error

        log.info(
            "Loaded %d parameters per node into %d nodes", len(node_params), num_nodes
        )

    def _initialize_node_states(self) -> None:
        """Add initial_values to a DAG from the config file"""

        log.info("Initial values: %s", self.initial_values)

        for node_id in self.network.nodes:
            self.network.nodes[node_id][
                "current_states"
            ] = self.initial_values.to_array()

    def _initialize_time(self) -> None:
        """Load mandatory time fields and start the clock."""
        try:
            start_time: datetime = self.config["start_time"]
            end_time: datetime = self.config["end_time"]
            time_step = self.config["time_step"]
        except KeyError as exc:
            log.critical("Config file is missing a required key: %s", exc)
            return  # Exit early if required keys are missing

        try:
            self.config["start_time"] = start_time.timestamp()
            log.info("start_time: %s (%s)", self.config["start_time"], start_time)
        except AttributeError:
            log.critical(
                "Wrong value for 'start_time' (expected format: YYYY-MM-DD hh:mm:ss)"
            )
            return  # Exit early on invalid start_time

        try:
            self.config["end_time"] = end_time.timestamp()
            log.info("end_time: %s (%s)", self.config["end_time"], end_time)
        except AttributeError:
            log.critical(
                "Wrong value for 'end_time' (expected format: YYYY-MM-DD hh:mm:ss)"
            )
            return  # Exit early on invalid end_time

        try:
            self.config["time_step"] = float(self.config["time_step"])
        except (TypeError, ValueError):
            log.critical("Wrong value for 'time_step' (expected type: float)")
            return  # Exit early on invalid time_step

        self.current_time = self.config["start_time"]  # start the clock

    def _get_forcing_filename(self, key: str) -> str:
        """Get the forcing file name from the config."""
        try:
            forcings = self.config["forcings"]
        except KeyError:
            log.critical("Expected a `forcings` key in the `config` dict")
            raise

        try:
            return forcings[key]
        except KeyError:
            log.critical("Expected a `%s` key in the `forcings` dict", key)
            raise
        except TypeError:
            log.critical(
                "Mismatched types for `forcings` (expected: dict, found: %s)",
                type(forcings).__name__,
            )
            raise

    def _read_csv(self, forcing_csv) -> pd.DataFrame:
        """Get the forcing file name from the config."""
        # Read the CSV file into a DataFrame
        try:
            forcing_df = pd.read_csv(
                forcing_csv,
                parse_dates=True,
            )
        except (FileNotFoundError, PermissionError) as e:
            log.critical(e)
            raise
        except pd.errors.EmptyDataError:
            log.critical("Found no columns to parse in `%s`", forcing_csv)
            raise  # Add this line!

        # Set the "time" column as index
        try:
            forcing_df.set_index("time", inplace=True)
        except KeyError:
            log.critical("Forcing files must have a `time` column")
            raise

        # Convert the "index" column into timestamps
        try:
            forcing_df.index = pd.Index(
                [
                    dt.timestamp()
                    for dt in pd.to_datetime(forcing_df.index, format="ISO8601")
                ]
            )
        except ValueError:
            log.critical("Data in the `time` column must be in the ISO 8601 format")
            raise

        return forcing_df

    def _load_forcing_series(self) -> None:
        """Get forcing series for each node."""
        for key in self.forcings_data.__dict__:
            forcing_csv = self._get_forcing_filename(key)
            series = self._read_csv(forcing_csv)
            setattr(self.forcings_data, key, series)

    def get_input(self, variable: pd.DataFrame, node_id: int, t: float) -> float:
        """Get a variable value at a given time and node_id."""
        return np.array(variable[str(node_id)].asof(t)).item()

    def _get_current_forcings(self) -> None:
        """Get forcings at current time for each node."""
        forcings = {}
        for key in self.forcings_data.__dict__:
            forcing_data = getattr(self.forcings_data, key)
            timestamps = forcing_data.index

            # Check if self.current_time is in the index
            if self.current_time in timestamps:
                forcings[key] = forcing_data.loc[self.current_time]
            else:
                # Use searchsorted to find the nearest indices
                idx = timestamps.searchsorted(self.current_time)

                # Handle edge cases: current_time is outside the range of the data
                if idx == 0:  # current_time is before the first timestamp
                    forcings[key] = forcing_data.iloc[0]
                elif idx == len(timestamps):  # current_time is after the last timestamp
                    forcings[key] = forcing_data.iloc[-1]
                else:
                    # Get the two closest times: one before and one after
                    time_before = timestamps[idx - 1]
                    time_after = timestamps[idx]

                    # Get values at the two nearest times
                    val_before = forcing_data.loc[time_before]
                    val_after = forcing_data.loc[time_after]

                    # Perform linear interpolation between the two points (timestamps)
                    delta_time = self.current_time - timestamps[idx - 1]
                    delta_total = timestamps[idx] - timestamps[idx - 1]

                    # Linear interpolation formula
                    interpolated_value = val_before + (val_after - val_before) * (
                        delta_time / delta_total
                    )

                    forcings[key] = interpolated_value

        current_forcings = pd.DataFrame(forcings)
        current_forcings.index = current_forcings.index.astype(int)
        target_order = self.get_nodes_in_reverse_topological_order()
        df_reordered = current_forcings.loc[target_order]  # .reset_index(drop=True)
        current_forcings_array = df_reordered.to_numpy()

        if self.current_forcings.size == 0:
            self.current_forcings = current_forcings_array
        else:
            self.current_forcings[:] = current_forcings_array

    def _get_current_states(self) -> None:
        """Get states at current time for each node."""

        node_ids = self.get_nodes_in_reverse_topological_order()

        current_states = []

        for node_id in node_ids:
            node = self.network.nodes[node_id]
            current_states.append(node["current_states"])

        if self.current_states.size == 0:
            self.current_states = np.array(current_states)
        else:
            self.current_states[:] = np.array(current_states)

    def startup(self, config_file: str) -> None:
        """Initialize HLM."""
        self._load_config(config_file)
        self._load_model_parameters()
        self._load_network_topology()
        self._load_network_parameters()
        self._load_initial_values_from_config()
        self._initialize_node_states()
        self._get_current_states()
        self._calculate_additional_network_parameters()
        self._load_forcing_series()
        self._initialize_time()
        self._get_current_forcings()  # to be able to get forcings at time 0 before the first update

    @abstractmethod
    def equations(
        self, t: float, y: NDArray[np.float64], node_id: int
    ) -> NDArray[np.float64]:
        """Compute the updated derivatives."""

    @abstractmethod
    def _calculate_additional_network_parameters(self) -> None:
        """Compute model-specific node parameters."""

    def get_nodes_in_reverse_topological_order(self) -> list[int]:
        """Get a list of node_ids in reverse topological order."""
        return list(nx.topological_sort(self.network))[::-1]

    def _solve_ivp_node(self, node_id: int, target_time: float) -> None:
        """Runs solve_ivp() for a node, from self.current_time to target_time"""

        def equations_wrapper(t, y):
            """Wraps self.equations() to match the signature expected by the integrator."""
            return self.equations(t, y, node_id)

        t_eval = np.arange(
            self.current_time,
            target_time + self.config["time_step"],
            self.config["time_step"],
        )

        ode_result = spi.solve_ivp(
            equations_wrapper,
            t_span=[self.current_time, target_time],
            t_eval=t_eval,
            y0=self.network.nodes[node_id]["current_states"],
            method=self.config["ode_solver"]["method"],
            atol=float(self.config["ode_solver"]["atol"]),
            rtol=float(self.config["ode_solver"]["rtol"]),
            max_step=float(self.config["ode_solver"]["max_step"]),
            dense_output=True,
        )

        if not ode_result.success:
            log.warning("Node ID #%s: %s", node_id, ode_result.message)

        ode_result.y[ode_result.y < 1e-12] = 0.0  # OutputConstraints_Model256_Hdf5

        self.network.nodes[node_id]["ode_result"] = ode_result

        self._update_current_states(node_id)
        self._stack_results(node_id)

    def _update_current_states(self, node_id: int):
        """Set node.current_states."""
        node = self.network.nodes[node_id]
        node["current_states"] = node["ode_result"].y[:, -1]

    def _stack_results(self, node_id: int):
        """Stack ode_results for a node."""
        node = self.network.nodes[node_id]
        ode_result = node["ode_result"]

        if "ode_result_t" not in node:
            node["ode_result_t"] = ode_result.t
        else:
            node["ode_result_t"] = np.hstack(
                (node["ode_result_t"], ode_result.t),
            )

        if "ode_result_y" not in node:
            node["ode_result_y"] = ode_result.y
        else:
            node["ode_result_y"] = np.hstack((node["ode_result_y"], ode_result.y))

    def _run_until(self, end_time: float) -> None:
        """Advance model state until the given time."""
        node_ids = self.get_nodes_in_reverse_topological_order()

        dt = np.array(self.current_time).astype("datetime64[s]")
        for node_id in tqdm(node_ids, unit="node", desc=f"{dt}"):
            self._solve_ivp_node(node_id, end_time)

        self._get_current_states()
        self._get_current_forcings()  # to be able to get forcings at time 0 before the first update

        self.current_time = end_time

    def solve(self) -> None:
        """Advance model state until the end time."""
        self._run_until(self.config["end_time"])

    def get_children_solutions(
        self,
        node_id: int,
        t: float,
    ) -> list[tuple[NDArray[np.float64], float]]:
        """Get the children solutions and flow partitions for a node."""
        children = list(self.network.successors(node_id))

        children_solutions = []
        for child_id in children:
            child_ode_result = self.network.nodes[child_id]["ode_result"]

            interpolated = child_ode_result.sol(t)

            if child_id in self.flow_partitions:
                child_flow_partition = self.flow_partitions[child_id][node_id]
                log.info(
                    "Node #%d will get %g·Q from Node #%d",
                    node_id,
                    child_flow_partition,
                    child_id,
                )
            else:
                child_flow_partition = 1.0
            children_solutions.append((interpolated, child_flow_partition))
        return children_solutions



"""Adds BMI functions to the base Hlm class."""

from dataclasses import dataclass, field
from typing import Generic

import numpy as np
from bmipy import Bmi
from numpy.typing import NDArray

from .hlm import F, Hlm, P, S


@dataclass
class BmiFields:
    """Required fields for every BMI-compliant model."""

    component_name: str = ""
    input_var_names: tuple[str, ...] = ()
    input_var_units: tuple[str, ...] = ()
    output_var_names: tuple[str, ...] = ()
    output_var_units: tuple[str, ...] = ()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not cls.component_name:
            raise NotImplementedError(f"{cls.__name__} must define 'component_name'")
        if not cls.input_var_names:
            raise NotImplementedError(f"{cls.__name__} must define 'input_var_names'")
        if not cls.input_var_units:
            raise NotImplementedError(f"{cls.__name__} must define 'input_var_units'")
        if not cls.output_var_names:
            raise NotImplementedError(f"{cls.__name__} must define 'output_var_names'")
        if not cls.output_var_units:
            raise NotImplementedError(f"{cls.__name__} must define 'output_var_units'")


@dataclass
class HlmBmi(
    Hlm[P, F, S], Bmi, Generic[P, F, S]
):  # pylint: disable=too-many-public-methods
    """Extend the Hlm base class to make it BMI-compliant."""

    bmi_fields: BmiFields = field(default_factory=BmiFields)

    # Model control functions

    def initialize(self, config_file: str) -> None:  # DONE
        """Perform startup tasks for the model."""
        self.startup(config_file)

    def update(self) -> None:  # DONE
        """Advance model state by one time step."""
        self._run_until(self.current_time + self.config["time_step"])

    def update_until(self, time: float) -> None:  # DONE
        """Advance model state until the given time."""
        self._run_until(time)

    def finalize(self) -> None:
        """Perform tear-down tasks for the model."""
        # TODO

    # Model information functions

    def get_component_name(self) -> str:  # DONE
        """Name of the component."""
        return self.bmi_fields.component_name

    def get_input_item_count(self) -> int:  # DONE
        """Count of a model's input variables."""
        return len(self.bmi_fields.input_var_names)

    def get_output_item_count(self) -> int:  # DONE
        """Count of a model's output variables."""
        return len(self.bmi_fields.output_var_names)

    def get_input_var_names(self) -> tuple[str, ...]:  # type: ignore
        # https://github.com/csdms/bmi-python/blob/b0ef8b640aee4b499c5d7a8b3b30d625c28a2194/src/bmipy/bmi.py#L110
        """Get the model's input variables."""
        return self.bmi_fields.input_var_names

    def get_output_var_names(self) -> tuple[str, ...]:  # type: ignore
        # https://github.com/csdms/bmi-python/blob/b0ef8b640aee4b499c5d7a8b3b30d625c28a2194/src/bmipy/bmi.py#L133
        """Get the model's output variables."""
        return self.bmi_fields.output_var_names

    # Variable information functions

    def get_var_grid(self, name: str) -> int:  # DONE
        """Get grid identifier for the given variable."""
        var_names = self.get_input_var_names() + self.get_output_var_names()
        for var_idx, var_name in enumerate(var_names):
            if var_name == name:
                return var_idx
        raise KeyError

    def get_var_type(self, name: str) -> str:  # DONE
        """Get data type of the given variable."""
        return str(self.get_value_ptr(name).dtype)

    def get_var_units(self, name: str) -> str:  # DONE
        """Get units of the given variable."""
        return (self.bmi_fields.input_var_units + self.bmi_fields.output_var_units)[
            self.get_var_grid(name)
        ]

    def get_var_itemsize(self, name: str) -> int:  # DONE
        """Get memory use for each array element in bytes."""
        return np.dtype(self.get_var_type(name)).itemsize

    def get_var_nbytes(self, name: str) -> int:  # DONE
        """Get size, in bytes, of the given variable."""
        return self.get_value_ptr(name).nbytes

    def get_var_location(self, name: str) -> str:
        """Get the grid element type that the a given variable is defined on."""
        return "node"  # FIXME: *node*: A point that has a coordinate pair or triplet

    # Time functions

    def get_current_time(self) -> float:
        """Return the current time of the model."""
        return self.current_time

    def get_start_time(self) -> float:
        """Start time of the model."""
        return self.config["start_time"]

    def get_end_time(self) -> float:
        """End time of the model."""
        return self.config["end_time"]

    def get_time_units(self) -> str:
        """Time units of the model."""
        return "s"

    def get_time_step(self) -> float:
        """Return the current time step of the model."""
        return self.config["time_step"]

    # Variable getter and setter functions

    def get_value(
        self, name: str, dest: NDArray[np.float64]
    ) -> NDArray[np.float64]:  # DONE
        """Get a copy of values of the given variable."""
        dest[:] = self.get_value_ptr(name).flatten()
        return dest

    def get_value_ptr(self, name: str) -> NDArray[np.float64]:  # DONE
        """Get a reference to values of the given variable."""

        for idx, var_name in enumerate(self.get_input_var_names()):
            if var_name == name:
                return self.current_forcings[:, idx]

        for idx, var_name in enumerate(self.get_output_var_names()):
            if var_name == name:
                return self.current_states[:, idx]

        raise KeyError

    def get_value_at_indices(
        self, name: str, dest: NDArray[np.float64], inds: NDArray[np.int_]
    ) -> NDArray[np.float64]:  # DONE
        """Get values at particular indices."""
        dest[:] = self.get_value_ptr(name)[inds]
        return dest

    def set_value(self, name: str, src: NDArray[np.float64]) -> None:  # DONE
        """Specify a new value for a model variable."""

        for idx, var_name in enumerate(self.get_input_var_names()):
            if var_name == name:
                self.current_forcings[:, idx] = src[:]
                return

        for idx, var_name in enumerate(self.get_output_var_names()):
            if var_name == name:
                self.current_states[:, idx] = src[:]
                return

        raise KeyError

    def set_value_at_indices(
        self, name: str, inds: NDArray[np.int_], src: NDArray[np.float64]
    ) -> None:
        """Specify a new value for a model variable at particular indices."""
        self.get_value_ptr(name)[inds] = src

    # # Grid information

    def get_grid_rank(self, grid: int) -> int:
        """Get number of dimensions of the computational grid."""
        raise NotImplementedError("`get_grid_rank` is not implemented")

    def get_grid_size(self, grid: int) -> int:
        """Get the total number of elements in the computational grid."""
        raise NotImplementedError("`get_grid_size` is not implemented")

    def get_grid_type(self, grid: int) -> str:
        """Get the grid type as a string."""
        return "unstructured"

    # Uniform rectilinear

    def get_grid_shape(self, grid: int, shape: NDArray[np.int_]) -> NDArray[np.int_]:
        """Get dimensions of the computational grid."""
        raise NotImplementedError("`get_grid_shape` is not implemented")

    def get_grid_spacing(
        self, grid: int, spacing: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Get distance between nodes of the computational grid."""
        raise NotImplementedError("`get_grid_spacing` is not implemented")

    def get_grid_origin(
        self, grid: int, origin: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Get coordinates for the lower-left corner of the computational grid."""
        raise NotImplementedError("`get_grid_origin` is not implemented")

    # Non-uniform rectilinear, curvilinear

    def get_grid_x(self, grid: int, x: NDArray[np.float64]) -> NDArray[np.float64]:
        """Get coordinates of grid nodes in the x direction."""
        raise NotImplementedError("`get_grid_x` is not implemented")

    def get_grid_y(self, grid: int, y: NDArray[np.float64]) -> NDArray[np.float64]:
        """Get coordinates of grid nodes in the y direction."""
        raise NotImplementedError("`get_grid_y` is not implemented")

    def get_grid_z(self, grid: int, z: NDArray[np.float64]) -> NDArray[np.float64]:
        """Get coordinates of grid nodes in the z direction."""
        raise NotImplementedError("`get_grid_z` is not implemented")

    def get_grid_node_count(self, grid: int) -> int:
        """Get the number of nodes in the grid."""
        return self.network.number_of_nodes()

    def get_grid_edge_count(self, grid: int) -> int:
        """Get the number of edges in the grid."""
        return self.network.number_of_edges()

    def get_grid_face_count(self, grid: int) -> int:
        """Get the number of faces in the grid."""
        raise NotImplementedError("`get_grid_face_count` is not implemented")

    def get_grid_edge_nodes(
        self, grid: int, edge_nodes: NDArray[np.int64]
    ) -> NDArray[np.int64]:
        """Get the edge-node connectivity."""
        edge_nodes[:] = np.array(
            [node for edge in self.network.edges for node in reversed(edge)],
        )
        return edge_nodes

    def get_grid_face_edges(
        self, grid: int, face_edges: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        """Get the face-edge connectivity."""
        raise NotImplementedError("`get_grid_face_edges` is not implemented")

    def get_grid_face_nodes(
        self, grid: int, face_nodes: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        """Get the face-node connectivity."""
        raise NotImplementedError("`get_grid_face_nodes` is not implemented")

    def get_grid_nodes_per_face(
        self, grid: int, nodes_per_face: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        """Get the number of nodes for each face."""
        raise NotImplementedError("`get_grid_nodes_per_face` is not implemented")
